@@ -1,14 +1,13 @@
-#!/usr/bin/env python
 # coding: utf-8
 
 """
-A MANAGESIEVE client
+A MANAGESIEVE client.
 
 A protocol for securely managing Sieve scripts on a remote server.
 This protocol allows a user to have multiple scripts, and also alerts
 a user to syntactically flawed scripts.
 
-Implementation based on <draft-martin-managesieve-12>.
+Implementation based on RFC 5804.
 """
 from __future__ import print_function
 
@@ -21,14 +20,16 @@ from future.utils import python_2_unicode_compatible
 import six
 
 from .digest_md5 import DigestMD5
+from . import tools
 
-CRLF = '\r\n'
 
-KNOWN_CAPABILITIES = ["IMPLEMENTATION", "SASL", "SIEVE",
-                      "STARTTLS", "NOTIFY", "LANGUAGE",
-                      "RENAME"]
+CRLF = b"\r\n"
 
-SUPPORTED_AUTH_MECHS = ["DIGEST-MD5", "PLAIN", "LOGIN"]
+KNOWN_CAPABILITIES = [u"IMPLEMENTATION", u"SASL", u"SIEVE",
+                      u"STARTTLS", u"NOTIFY", u"LANGUAGE",
+                      u"VERSION"]
+
+SUPPORTED_AUTH_MECHS = [u"DIGEST-MD5", u"PLAIN", u"LOGIN"]
 
 
 class Error(Exception):
@@ -79,15 +80,15 @@ class Client(object):
         self.srvport = srvport
         self.__debug = debug
         self.sock = None
-        self.__read_buffer = ""
+        self.__read_buffer = b""
         self.authenticated = False
         self.errcode = None
 
         self.__capabilities = {}
-        self.__respcode_expr = re.compile(r"(OK|NO|BYE)\s*(.+)?")
-        self.__error_expr = re.compile(r'(\([\w/-]+\))?\s*(".+")')
-        self.__size_expr = re.compile(r"\{(\d+)\+?\}")
-        self.__active_expr = re.compile("ACTIVE", re.IGNORECASE)
+        self.__respcode_expr = re.compile(br"(OK|NO|BYE)\s*(.+)?")
+        self.__error_expr = re.compile(br'(\([\w/-]+\))?\s*(".+")')
+        self.__size_expr = re.compile(br"\{(\d+)\+?\}")
+        self.__active_expr = re.compile(br"ACTIVE", re.IGNORECASE)
 
     def __del__(self):
         if self.sock is not None:
@@ -114,7 +115,7 @@ class Client(object):
         :rtype: string
         :returns: the read block (can be empty)
         """
-        buf = ""
+        buf = b""
         if len(self.__read_buffer):
             limit = (
                 size if size <= len(self.__read_buffer) else
@@ -129,6 +130,7 @@ class Client(object):
             buf += self.sock.recv(size)
         except (socket.timeout, ssl.SSLError):
             raise Error("Failed to read %d bytes from the server" % size)
+        self.__dprint(buf)
         return buf
 
     def __read_line(self):
@@ -146,17 +148,18 @@ class Client(object):
         :rtype: string
         :return: the read line
         """
-        ret = ""
+        ret = b""
         while True:
             try:
                 pos = self.__read_buffer.index(CRLF)
-                ret = self.__read_buffer[0:pos]
+                ret = self.__read_buffer[:pos]
                 self.__read_buffer = self.__read_buffer[pos + len(CRLF):]
                 break
             except ValueError:
                 pass
             try:
                 nval = self.sock.recv(self.read_size)
+                self.__dprint(nval)
                 if not len(nval):
                     break
                 self.__read_buffer += nval
@@ -170,9 +173,9 @@ class Client(object):
 
             m = self.__respcode_expr.match(ret)
             if m:
-                if m.group(1) == "BYE":
+                if m.group(1) == b"BYE":
                     raise Error("Connection closed by server")
-                if m.group(1) == "NO":
+                if m.group(1) == b"NO":
                     self.__parse_error(m.group(2))
                 raise Response(m.group(1), m.group(2))
         return ret
@@ -190,7 +193,7 @@ class Client(object):
         :return: a tuple of the form (code, data, response). If
         nblines is provided, code and data can be equal to None.
         """
-        resp, code, data = ("", None, None)
+        resp, code, data = (b"", None, None)
         cpt = 0
         while True:
             try:
@@ -201,6 +204,8 @@ class Client(object):
                 break
             except Literal as inst:
                 resp += self.__read_block(inst.value)
+                if not resp.endswith(CRLF):
+                    resp += self.__read_line() + CRLF
                 continue
             if not len(line):
                 continue
@@ -222,14 +227,18 @@ class Client(object):
         """
         ret = []
         for a in args:
-            if type(a) in [str, unicode] and self.__size_expr.match(a) is None:
-                ret += ['"%s"' % a.encode('utf-8')]
+            if isinstance(a, six.binary_type):
+                if self.__size_expr.match(a):
+                    ret += [a]
+                else:
+                    ret += [b'"' + a + b'"']
                 continue
-            ret += ["%s" % str(a)]
+            ret += [bytes(str(a).encode("utf-8"))]
         return ret
 
     def __send_command(
-            self, name, args=[], withcontent=False, extralines=[], nblines=-1):
+            self, name, args=None, withcontent=False, extralines=None,
+            nblines=-1):
         """Send a command to the server.
 
         If args is not empty, we concatenate the given command with
@@ -249,15 +258,20 @@ class Client(object):
         :returns: a tuple of the form (code, data[, response])
 
         """
-        tosend = name
-        if len(args):
-            tosend += " " + " ".join(self.__prepare_args(args))
-        self.__dprint("Command: %s" % tosend)
-        self.sock.sendall("%s%s" % (tosend, CRLF))
-        for l in extralines:
-            self.sock.sendall("%s%s" % (l, CRLF))
+        tosend = name.encode("utf-8")
+        if args:
+            tosend += b" " + b" ".join(self.__prepare_args(args))
+        self.__dprint(b"Command: " + tosend)
+        self.sock.sendall(tosend + CRLF)
+        if extralines:
+            for l in extralines:
+                self.sock.sendall(l + CRLF)
         code, data, content = self.__read_response(nblines)
 
+        if isinstance(code, six.binary_type):
+            code = code.decode("utf-8")
+        data = data.decode("utf-8")
+
         if withcontent:
             return (code, data, content)
         return (code, data)
@@ -269,11 +283,13 @@ class Client(object):
 
         for l in capabilities.splitlines():
             parts = l.split(None, 1)
-            cname = parts[0].strip('"')
+            cname = parts[0].strip(b'"').decode("utf-8")
             if cname not in KNOWN_CAPABILITIES:
                 continue
-            self.__capabilities[cname] = \
-                parts[1].strip('"') if len(parts) > 1 else None
+            self.__capabilities[cname] = (
+                parts[1].strip(b'"').decode("utf-8")
+                if len(parts) > 1 else None
+            )
         return True
 
     def __parse_error(self, text):
@@ -291,7 +307,7 @@ class Client(object):
         """
         m = self.__size_expr.match(text)
         if m is not None:
-            self.errcode = ""
+            self.errcode = b""
             self.errmsg = self.__read_block(int(m.group(1)) + 2)
             return
 
@@ -299,12 +315,12 @@ class Client(object):
         if m is None:
             raise Error("Bad error message")
         if m.group(1) is not None:
-            self.errcode = m.group(1).strip("()")
+            self.errcode = m.group(1).strip(b"()")
         else:
-            self.errcode = ""
-        self.errmsg = m.group(2).strip('"')
+            self.errcode = b""
+        self.errmsg = m.group(2).strip(b'"')
 
-    def _plain_authentication(self, login, password, authz_id=""):
+    def _plain_authentication(self, login, password, authz_id=b""):
         """SASL PLAIN authentication
 
         :param login: username
@@ -315,8 +331,8 @@ class Client(object):
             login = login.encode("utf-8")
         if isinstance(login, six.text_type):
             password = password.encode("utf-8")
-        params = base64.b64encode('\0'.join([authz_id, login, password]))
-        code, data = self.__send_command("AUTHENTICATE", ["PLAIN", params])
+        params = base64.b64encode(b'\0'.join([authz_id, login, password]))
+        code, data = self.__send_command("AUTHENTICATE", [b"PLAIN", params])
         if code == "OK":
             return True
         return False
@@ -328,9 +344,9 @@ class Client(object):
         :param password: clear password
         :return: True on success, False otherwise.
         """
-        extralines = ['"%s"' % base64.b64encode(login),
-                      '"%s"' % base64.b64encode(password)]
-        code, data = self.__send_command("AUTHENTICATE", ["LOGIN"],
+        extralines = [b'"%s"' % base64.b64encode(login.encode("utf-8")),
+                      b'"%s"' % base64.b64encode(password.encode("utf-8"))]
+        code, data = self.__send_command("AUTHENTICATE", [b"LOGIN"],
                                          extralines=extralines)
         if code == "OK":
             return True
@@ -344,7 +360,7 @@ class Client(object):
         :return: True on success, False otherwise.
         """
         code, data, challenge = \
-            self.__send_command("AUTHENTICATE", ["DIGEST-MD5"],
+            self.__send_command("AUTHENTICATE", [b"DIGEST-MD5"],
                                 withcontent=True, nblines=1)
         dmd5 = DigestMD5(challenge, "sieve/%s" % self.srvaddr)
 
@@ -362,7 +378,7 @@ class Client(object):
             return True
         return False
 
-    def __authenticate(self, login, password, authz_id="", authmech=None):
+    def __authenticate(self, login, password, authz_id=b"", authmech=None):
         """AUTHENTICATE command
 
         Actually, it is just a wrapper to the real commands (one by
@@ -397,7 +413,7 @@ class Client(object):
                 return True
             return False
 
-        self.errmsg = "No suitable mechanism found"
+        self.errmsg = b"No suitable mechanism found"
         return False
 
     def __starttls(self, keyfile=None, certfile=None):
@@ -461,12 +477,13 @@ class Client(object):
 
         :rtype: string
         """
-        if type(self.__capabilities["SIEVE"]) == str:
+        if isinstance(self.__capabilities["SIEVE"], six.string_types):
             self.__capabilities["SIEVE"] = self.__capabilities["SIEVE"].split()
         return self.__capabilities["SIEVE"]
 
     def connect(
-            self, login, password, authz_id="", starttls=False, authmech=None):
+            self, login, password, authz_id=b"", starttls=False,
+            authmech=None):
         """Establish a connection with the server.
 
         This function must be used. It read the server capabilities
@@ -508,8 +525,8 @@ class Client(object):
 
         :rtype: string
         """
-        code, data, capabilities = \
-            self.__send_command("CAPABILITY", withcontent=True)
+        code, data, capabilities = (
+            self.__send_command("CAPABILITY", withcontent=True))
         if code == "OK":
             return capabilities
         return None
@@ -524,7 +541,8 @@ class Client(object):
         :param scriptsize: script's size
         :rtype: boolean
         """
-        code, data = self.__send_command("HAVESPACE", [scriptname, scriptsize])
+        code, data = self.__send_command(
+            "HAVESPACE", [scriptname.encode("utf-8"), scriptsize])
         if code == "OK":
             return True
         return False
@@ -546,14 +564,15 @@ class Client(object):
         for l in listing.splitlines():
             if self.__size_expr.match(l):
                 continue
-            m = re.match(r'"([^"]+)"\s*(.+)', l)
+            m = re.match(br'"([^"]+)"\s*(.+)', l)
             if m is None:
-                ret += [l.strip('"')]
-            else:
-                if self.__active_expr.match(m.group(2)):
-                    active_script = m.group(1)
-                else:
-                    ret += [m.group(1)]
+                ret += [l.strip(b'"').decode("utf-8")]
+                continue
+            script = m.group(1).decode("utf-8")
+            if self.__active_expr.match(m.group(2)):
+                active_script = script
+                continue
+            ret += [script]
         self.__dprint(ret)
         return (active_script, ret)
 
@@ -568,12 +587,12 @@ class Client(object):
         :returns: the script's content on succes, None otherwise
         """
         code, data, content = self.__send_command(
-            "GETSCRIPT", [name], withcontent=True)
+            "GETSCRIPT", [name.encode("utf-8")], withcontent=True)
         if code == "OK":
             lines = content.splitlines()
             if self.__size_expr.match(lines[0]) is not None:
                 lines = lines[1:]
-            return "\n".join(lines)
+            return u"\n".join([line.decode("utf-8") for line in lines])
         return None
 
     @authentication_required
@@ -586,12 +605,10 @@ class Client(object):
         :param content: script's content
         :rtype: boolean
         """
-        if type(content) is unicode:
-            content = content.encode("utf-8")
-
-        content = "{%d+}%s%s" % (len(content), CRLF, content)
-        code, data = \
-            self.__send_command("PUTSCRIPT", [name, content])
+        content = tools.to_bytes(content)
+        content = tools.to_bytes("{%d+}" % len(content)) + CRLF + content
+        code, data = (
+            self.__send_command("PUTSCRIPT", [name.encode("utf-8"), content]))
         if code == "OK":
             return True
         return False
@@ -605,7 +622,8 @@ class Client(object):
         :param name: script's name
         :rtype: boolean
         """
-        code, data = self.__send_command("DELETESCRIPT", [name])
+        code, data = self.__send_command(
+            "DELETESCRIPT", [name.encode("utf-8")])
         if code == "OK":
             return True
         return False
@@ -623,19 +641,24 @@ class Client(object):
         :param newname: new script's name
         :rtype: boolean
         """
-        if "RENAMESCRIPT" in self.__capabilities:
+        if "VERSION" in self.__capabilities:
             code, data = self.__send_command(
-                "RENAMESCRIPT", [oldname, newname])
+                "RENAMESCRIPT",
+                [oldname.encode("utf-8"), newname.encode("utf-8")])
             if code == "OK":
                 return True
             return False
 
         (active_script, scripts) = self.listscripts()
-        if scripts is None or oldname not in scripts:
-            self.errmsg = "Old script does not exist"
+        condition = (
+            oldname != active_script and
+            (scripts is None or oldname not in scripts)
+        )
+        if condition:
+            self.errmsg = b"Old script does not exist"
             return False
         if newname in scripts:
-            self.errmsg = "New script already exists"
+            self.errmsg = b"New script already exists"
             return False
         oldscript = self.getscript(oldname)
         if oldscript is None:
@@ -661,7 +684,8 @@ class Client(object):
         :param scriptname: script's name
         :rtype: boolean
         """
-        code, data = self.__send_command("SETACTIVE", [scriptname])
+        code, data = self.__send_command(
+            "SETACTIVE", [scriptname.encode("utf-8")])
         if code == "OK":
             return True
         return False
@@ -675,12 +699,12 @@ class Client(object):
         :param name: script's content
         :rtype: boolean
         """
-        if type(content) is unicode:
-            content = content.encode("utf-8")
-
-        content = "{%d+}%s%s" % (len(content), CRLF, content)
-        code, data = \
-            self.__send_command("CHECKSCRIPT", [content])
+        if "VERSION" not in self.__capabilities:
+            raise NotImplementedError(
+                "server does not support CHECKSCRIPT command")
+        content = tools.to_bytes(
+            u"{%d+}%s%s" % (len(content), str(CRLF), content))
+        code, data = self.__send_command("CHECKSCRIPT", [content])
         if code == "OK":
             return True
         return False
